// Copyright (c) 1995, 1996 Regents of the University of California.
// All rights reserved.
//
// This software was developed by the Arcadia project
// at the University of California, Irvine.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the University of California, Irvine.  The name of the
// University may not be used to endorse or promote products derived
// from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

// File: ArcPerspective.java
// Classes: ArcPerspective
// Original Author: Jason Robbins
// $Id: ArcPerspective.java,v 1.6 1997/01/08 20:16:49 jrobbins Exp $

package uci.graphedit;

import java.util.*;
import java.awt.*;
import java.io.*;



/** A DiagramElement that draws arcs between ports. This version
 * automatically routes a rectilinear arc. The routing is not very
 * good. It avoids the source and sink nodes and no other nodes. It is
 * basically case-analysis, and some of the cases are wrong or
 * missing. */

public class ArcPerspective extends DiagramElement {
  
  protected Fig _sourcePortFig;
  protected Fig _destPortFig;
  protected Perspective _sourcePerspective;
  protected Perspective _destPerspective;

  public ArcPerspective(Fig s, Fig d,
			Perspective sp, Perspective dp,
			NetArc a) {
    _sourcePortFig = s;
    _destPortFig = d;
    _sourcePerspective = sp;
    _destPerspective = dp;
    Rectangle bbox = getBoundingBox();
    position(bbox.x, bbox.y);
    owner(a);
  }

  /** Draw this object.
   * @see DiagramElement#drawSelected */
  public void draw(Graphics g) {
    g.setColor(Color.black);
    drawArc(g);
  }

  /** Draw this object when the user has selected it.
   * @see DiagramElement#drawSelected */
  public void drawSelected(Graphics g) {
    /* needs-more-work: real handles that can be dragged */
    g.setColor(Globals.prefs().handleColor());
    drawArc(g);
  }

  /** reply the bounding box for this arc.
   * @see DiagramElement#getBoundingBox */
  public Rectangle getBoundingBox() {
    if (rects.size() == 0) computeRoute();
    Rectangle bbox = null;
    Enumeration curRect = rects.elements();
    while (curRect.hasMoreElements()) {
      Rectangle r = (Rectangle) curRect.nextElement();
      if (bbox == null) { bbox = new Rectangle(r.x, r.y, r.width, r.height); }
      else { bbox.add(r); }
    }
    if (bbox == null) {
      bbox = _sourcePortFig.getBoundingBox();
      bbox.add(_destPortFig.getBoundingBox());
    }
    return bbox;
  }

  /** reply true iff this arc contains the given point.
   * @see DiagramElement#inside
   */
  public boolean inside(int x, int y) {
    Enumeration curRect = rects.elements();
    while (curRect.hasMoreElements()) {
      Rectangle r = (Rectangle) curRect.nextElement();
      if (r.inside(x, y)) { return true;}
    }
    return false;
  }

  /** reply true iff this arc intersects the given rectangle.
   * @see DiagramElement#intersects
   */
  public boolean intersects(Rectangle rect) {
    Enumeration curRect = rects.elements();
    while (curRect.hasMoreElements()) {
      Rectangle r = (Rectangle) curRect.nextElement();
      if (r.intersects(rect)) { return true;}
    }
    return false;
  }

  public void dispose(Editor ed) {
    ((NetArc)owner()).dispose();
    super.dispose(ed);
  }

  void readInFile(DataInputStream in_file) throws IOException { }
  void writeToFile(DataOutputStream out_file) throws IOException { }


  /** This is my representation for rectilinear polygons. A polyline
  /** would be better, if it could be drawn with a thick line and if
  /** it was easy to detect if the mouse is near it */
  protected Vector rects = new Vector();
  
  /** inforation needed for routing */
  protected Point lastPoint;

  /** Add a routing point by calling addPoint(Point).
   * Called from computeRoute() */
  protected final void addPoint(int x, int y) { addPoint(new Point(x, y)); }

  /** Add a routing point. Called from computeRoute() */
  protected void addPoint(Point p) {
    if (lastPoint != null) {
      if (lastPoint.x != p.x && lastPoint.y != p.y) {
	addPoint(p.x, lastPoint.y);
	addPoint(p);
	return;
      }
      Rectangle r = new Rectangle(p);
      r.add(lastPoint);
      rects.addElement(r);
    }
    lastPoint = p;
  }

  /** After the routing has been computed, there might be some
   * post-processing needed to make the arc look nice. */
  protected void finishArc() {
    int W = 1; /* extra width to add to each rect */
    Enumeration curRect = rects.elements();
    while (curRect.hasMoreElements()) {
      Rectangle r = (Rectangle) curRect.nextElement();
      r.reshape(r.x-W, r.y-W, r.width+W*2, r.height+W*2);
    }
  }

  /** forget all old routing info and prepare to recalculate */
  protected void reset() { rects.removeAllElements(); lastPoint = null; }


  /** find the route that the arc should follow. Basically case
   * analysis to route around source and destination nodes. <p>
   * needs-more-work: a better algorithm would really be useful. */
  protected void computeRoute() {
    reset();
    Point srcPt = _sourcePortFig.center();
    Point dstPt = _destPortFig.center();
    Rectangle srcRR = _sourcePerspective.routingRect();
    Rectangle dstRR = _destPerspective.routingRect();

    Point mid = new Point( ((srcRR.x + srcRR.width/2) +
			   (dstRR.x + dstRR.width/2)) / 2,
		          ((srcRR.y + srcRR.height/2)+
			   (dstRR.y + dstRR.height/2)) /2);
    NetPort srcPort = (NetPort) _sourcePortFig.owner();
    NetPort dstPort = (NetPort) _destPortFig.owner();
    int srcSector = srcPort.getPortSector(_sourcePerspective);
    int dstSector = dstPort.getPortSector(_destPerspective);
    addPoint(srcPt);
    if ((srcSector==1) && (srcPt.y>dstPt.y)) {
      /* top to bottom */
      addPoint(srcPt.x,srcRR.y+srcRR.height);
      if (srcRR.x>dstRR.x) { /* move left and down */
	addPoint(srcRR.x,srcRR.y+srcRR.height);
	addPoint(srcRR.x,mid.y);
      } else {	/* move right and down */
	addPoint(srcRR.x+srcRR.width, srcRR.y+srcRR.height);
	addPoint(srcRR.x+srcRR.width,mid.y);
      }
    } 		/* left to right */
    else if ((srcSector==2) && (srcPt.x<dstPt.x)) {
      addPoint(srcRR.x,srcPt.y);
      if (srcRR.y>dstRR.y) {
	addPoint(srcRR.x,srcRR.y);
	addPoint(mid.x,srcRR.y);
      } else {
	addPoint(srcRR.x,srcRR.y+srcRR.height);
	addPoint(mid.x,srcRR.y+srcRR.height);
      }
    }
    else if ((srcSector==3) && (srcPt.y<dstPt.y)) {
      /* bottom to top */
      addPoint(srcPt.x,srcRR.y);
      if (srcRR.x>dstRR.x) {
	addPoint(srcRR.x,srcRR.y);
	addPoint(srcRR.x,mid.y);
      } else {
	addPoint(srcRR.x+srcRR.width, srcRR.y);
	addPoint(srcRR.x+srcRR.width,mid.y);
      }
    }
    else if ((srcSector==4) && (srcPt.x>dstPt.x)) {
      addPoint(srcRR.x+srcRR.width,srcPt.y);
      if (srcRR.y>dstRR.y) {
	addPoint(srcRR.x+srcRR.width, srcRR.y);
	addPoint(mid.x,srcRR.y);
      } else {
	addPoint(srcRR.x+srcRR.width,srcRR.y+srcRR.height);
	addPoint(mid.x,srcRR.y+srcRR.height);
      }
    }
    else {
      if ((srcSector==1) || (srcSector==3)) {
	addPoint(srcPt.x,mid.y);
      }
      else addPoint(mid.x,srcPt.y);
    }
    if ((dstSector==1) && (dstPt.y>srcPt.y)) {
      /* top to bottom */
      if (dstRR.x>srcRR.x) {
	/* move left and down */
	addPoint(dstRR.x,mid.y);
	addPoint(dstRR.x,dstRR.y+ dstRR.height);
      } else {	/* move right and down */
	addPoint(dstRR.x+dstRR.width,mid.y);
	addPoint(dstRR.x+dstRR.width, dstRR.y+dstRR.height);
      }
      addPoint(dstPt.x,dstRR.y+dstRR.height);
    } /* left to right */
    else if ((dstSector==2) && (dstPt.x<srcPt.x)) {
      if (dstRR.y>srcRR.y) {
	addPoint(mid.x,dstRR.y);
	addPoint(dstRR.x,dstRR.y);
      } else {
	addPoint(mid.x,dstRR.y+dstRR.height);
	addPoint(dstRR.x, dstRR.y+dstRR.height);
      }
      addPoint(dstRR.x,dstPt.y);
    }
    else if ((dstSector==3) && (dstPt.y<srcPt.y)) { /* bottom to top */
      if (dstRR.x>srcRR.x) {
	addPoint(dstRR.x,mid.y);
	addPoint(dstRR.x,dstRR.y);
      } else {
	addPoint(dstRR.x+dstRR.width,mid.y);
	addPoint(dstRR.x+dstRR.width, dstRR.y);
      }
      addPoint(dstPt.x,dstRR.y);
    }
    else if ((dstSector==4) && (dstPt.x>srcPt.x)) {
      if (dstRR.y>srcRR.y) {
	addPoint(mid.x,dstRR.y);
	addPoint(dstRR.x+dstRR.width, dstRR.y);
      } else {
	addPoint(mid.x,dstRR.y+dstRR.height);
	addPoint(dstRR.x+dstRR.width,
					   dstRR.y+dstRR.height);
      }
      addPoint(dstRR.x+dstRR.width,dstPt.y);
    }
    else {
      if ((dstSector==1) || (dstSector==3)) {
	addPoint(dstPt.x,mid.y);
      }
      else { addPoint(mid.x,dstPt.y); }
    }
    addPoint(dstPt);
    finishArc();
  } /* end computeRoute */

  /** Draw the arc */
  void drawArc(Graphics g) {
    computeRoute();
    Enumeration curRect = rects.elements();
    while (curRect.hasMoreElements()) {
      Rectangle r = (Rectangle)curRect.nextElement();
      g.fillRect(r.x, r.y, r.width, r.height);
    }
  }

  /** reply a new selection object that can be used to manipulate this
   * object. For now, use SelectionHandles. needs-more-work: users
   * should be able to adjust arcs by dragging on handles */  
  public Selection selectionObject() {
    return new SelectionHandles(this);
  }

  protected Object srcOwner(){ return _sourcePortFig.owner(); }

  protected Object dstOwner(){ return _destPortFig.owner();}

} /* end class ArcPerspective */

